import packaging.requirements
import packaging.specifiers
import pathlib
import io
import subprocess
import argparse
import typing


def clean_lines(f: io.TextIOWrapper) -> typing.Generator[str, None, None]:
    for line in f:
        # remove comments and hashes
        # because packaging.requirements.Requirement can't read hashes
        # and comments are useless here
        if not line.lstrip().startswith(("#", "--hash")):
            yield line.rstrip("\n").rstrip("\\")


def compare_requirements(
    file: pathlib.Path,
) -> tuple[
    packaging.requirements.Requirement,
    packaging.requirements.Requirement,
    packaging.requirements.Requirement,
]:
    set_before = set()
    set_after = set()
    with file.open() as f:
        set_after = {
            packaging.requirements.Requirement(line) for line in clean_lines(f)
        }

    set_before = {
        packaging.requirements.Requirement(line)
        for line in clean_lines(retrieve_requirements(file))
    }

    d_new = {s.name: s for s in set_after.difference(set_before)}
    d_old = {s.name: s for s in set_before.difference(set_after)}

    set_added = {v for k, v in d_new.items() if k not in d_old}
    set_updated = {(d_old[k], v) for k, v in d_new.items() if k in d_old}
    set_removed = {v for k, v in d_old.items() if k not in d_new}

    return (set_added, set_updated, set_removed)


def generate_changes(
    added: set[packaging.requirements.Requirement],
    updated: set[
        tuple[
            packaging.requirements.Requirement,
            packaging.requirements.Requirement,
        ]
    ],
    removed: set[packaging.requirements.Requirement],
) -> str:
    description = []
    for a in added:
        description.append(
            f"- Added `{a.name}` version "
            f"`{retrieve_equal_version(a.specifier).version}`"
        )
    for u in updated:
        description.append(
            f"- Bump `{u[0].name}` from "
            f"`{retrieve_equal_version(u[0].specifier).version}` to "
            f"`{retrieve_equal_version(u[1].specifier).version}`"
        )
    for r in removed:
        description.append(
            f"- Removed `{r.name}` version "
            f"`{retrieve_equal_version(r.specifier).version}`"
        )
    return "\r\n".join(description)


def retrieve_equal_version(
    specifier: packaging.specifiers.SpecifierSet,
) -> packaging.specifiers.Specifier:
    for s in specifier:
        if s.operator == "==":
            return s


def retrieve_requirements(file: pathlib.Path) -> io.StringIO:
    p = subprocess.run(
        ["git", "show", f"HEAD:{file.name}"],
        stdout=subprocess.PIPE,
    )
    content = io.StringIO(p.stdout.decode())
    return content


def parser() -> argparse.ArgumentParser:
    p = argparse.ArgumentParser()
    p.add_argument(
        "requirements_file",
        nargs="?",
        default=pathlib.Path("requirements.txt"),
        type=pathlib.Path,
    )
    p.add_argument(
        "-o",
        nargs=1,
        default=pathlib.Path("description.md"),
        type=pathlib.Path,
        required=False,
    )
    return p


def main():
    args = parser().parse_args()
    (added, updated, removed) = compare_requirements(args.requirements_file)
    args.o.write_text(generate_changes(added, updated, removed))


if __name__ == "__main__":
    main()
